'' Copyright (c) 2008 Philip C. Pilgrim
'' (See end of file for terms of use.)
''
''-----------[ Description ]----------------------------------------------------
''
'' This Propeller object provides buffered serial receive and unbuffered serial
'' transmit on either a single pin or a pair of pins. Pins can be either open
'' drain or bipolar-driven, mark-high (normal) or mark-low (inverted).
''
''-----------[ Revision History ]-----------------------------------------------
''
'' 2009.12.9: Added output break capability and method.
''
''-----------[ Program ]--------------------------------------------------------

CON

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000 

  rf_pin        = 3             'Input pin for RF from antenna or RF amplifier.
  fb_pin        = 4             'Feedback pin for delta-sigma ADC.
  if_pin        = 8             'I quadrature output pin.
  qf_pin        = 9             'Q quadrature output pin.
  au_pin        = 11            'Audio output pin.
  fsample       = 16000         'Audio sample rate in Hz.
  clks          = 80_000_000 / fsample 'Processor clocks per sample (assumes 80MHz clock).                 
  mclks         = clks / 2      'Clocks for half a sample ("zero" level for ADC).

OBJ

  tv : "tv_wtext"

VAR

  long  qStack[32]
  long  dStack[32]
  long  xStack[32]
  long  audio, frequency, sum, avg, dif, n
               
PUB Start

  tv.start(12)                                  'Start TV output to display receive frequency.
  dira[fb_pin]~~                                'Set the feedback pin to an output.
  ctra := %01001 << 26 | fb_pin << 9 | rf_pin   'Set up ctra for negative feedback.
  q_freq := @frequency                          'Initialize address for quadrature cog to find frequency.
  d_audio := a_audio := @audio                  'Initialize audtio addresses in two cogs.                 
  cognew(@quadrature, 0)                        'Start quadrature generator.
  cognew(@demod, 0)                             'Start the detector/demodulator.
  cognew(@audio_out, 0)                         'Start the audio output.
  repeat                                        'Scan a range of frequencies, displaying each one...
    repeat frequency from 14_000_000 to 14_050_000 step 500
      tv.out(0)
      tv.decn(frequency / 1000, 4)
      tv.out(".")
      tv.decn(frequency // 1000, -3)
      tv.str(string(" KHz "))
      waitcnt(cnt + clkfreq * 2)

DAT

'' Generate two square waves at the receive frequency, 90 degrees out of phase.

              org       0
              
quadrature    mov       dira,q_dira0            'Set the quadrature pins to output.
              
:query_lp     rdlong    freq,q_freq             'Get the target receive frequency.
              cmp       freq,pfreq wz           'Same as previous reading?
        if_z  jmp       #:query_lp              '  Yes: Nothing to do.

              mov       ra,freq                 'Save frequency in high part of ra:rb.
              mov       rb,#0
              mov       rx,#%0_00100_000        'Assume NCO output at first.
              mov       _90deg,_4000_0000       'Shift for NCO is 2**32 / 4
              cmp       ra,_4mhz wc             'Is frequency less than 4MHz?
        if_c  jmp       #:do_mode               '  Yes: Can't do PLL, so continue w/NCO.

:do_pll       mov       rx,#%0_00010_011        '  No:  Change mode to PLLx1.

:pll_lp       cmp       ra,_8mhz wc             'Is NCO frequency less than 8MHz?
        if_c  jmp       #:do_mode               '  Yes: Done adjusting.

              add       rx,#1                   'Bump PLL mode by one.
              shr       ra,#1 wc                'Divide NCO frequency by two.
              rcr       rb,#1
              shr       _90deg,#1               'Also divide quadrature adder by two.
              jmp       #:pll_lp                'Try again.

:do_mode      movi      q_ctra0,rx              'Save PLL mode in counter buffers.
              movi      q_ctrb0,rx        
                                                
              rdlong    rx,#0                   'Fetch clk_freq.
              call      #divabx                 'Divide into NCO freq (* 2**32)
              
              mov       ctra,q_ctra0            'Set new coutner modes to counters.
              mov       ctrb,q_ctrb0
              mov       frqa,#0                 'Stop counting momentarily.
              mov       frqb,#0
              mov       phsa,#0                 'Initialize one phase to zero,
              mov       phsb,_90deg             '  the other to the quadrature offset.
              add       phsb,rb                 'Then adjust the phase by 4xNCO freq
              add       phsb,rb                 '  to account for frqb being loaded  
              add       phsb,rb                 '  four clocks later than frqa.
              add       phsb,rb
              mov       frqa,rb                 'Now load the NCO frequencies.
              mov       frqb,rb
              mov       pfreq,freq              'Save the new frequency in pfreq.
              jmp       #:query_lp              'Done: back for a new frequency.

'-----------[ Divide 64 / 32 = 32 ]--------------------------------------------  

' rb := ra:rb / rx

divabx        mov       cctr,#32 wc
:loop         rcl       rb,#1 wc
              rcl       ra,#1 wc
        if_c  sub       ra,rx
       if_nc  cmpsub    ra,rx wc
              djnz      cctr,#:loop
              rcl       rb,#1 wc
divabx_ret    ret

'-----------[ Constants & Variables ]------------------------------------------


q_freq        long      0-0
q_ctra0       long      %00100 << 26 | if_pin
q_ctrb0       long      %00100 << 26 | qf_pin
q_dira0       long      1 << if_pin | 1 << qf_pin

freq          long      0
pfreq         long      0

_4000_0000    long      $4000_0000
_4mhz         long      4_000_000
_8mhz         long      8_000_000

ra            res       1
rb            res       1
rx            res       1
cctr          res       1
_90deg        res       1


''===================================================================

'' I and Q mixers and detector.

              org       0
demod         mov       frqa,#1                 'Initialize counters to count up by one.
              mov       frqb,#1
              mov       ctra,d_ctra0            'Set ctra to increment on I XOR delta-sigma feedback.
              mov       ctrb,d_ctrb0            'Set ctrb to increment on Q XOR delta-sigma feedback.
              mov       time,cnt                'Initialize sample interval.
              add       time,sample

:mix          waitcnt   time,sample             'Wait for next sample time.
              mov       i_samp,phsa             'Get the I sample
              mov       phsa,#0                 ' and zero it.
              mov       q_samp,phsb             'Get the Q sample
              mov       phsb,#0                 ' and zero it.
              mov       acc,i_samp              'Amplitude = sqrt(I*I + Q*Q).
              call      #square
              mov       i_samp,acc
              mov       acc,q_samp
              call      #square
              add       acc,i_samp
              call      #sqrt
              wrlong    acc,d_audio             'Write the amplitude to the audio long.
              jmp       #:mix                   'Back for another sample.
              
''--------[square]---------------------------------------------------

'' acc = (acc - half) ** 2   (Multiply routine from "Propeller Guts".)

square        sub       acc,half
              abs       acc,acc
              mov       ctr,#16
              mov       accx,acc
              shl       accx,#16
              shr       acc,#1 wc
              
:sqlp   if_c  add       acc,accx wc
              rcr       acc,#1 wc
              djnz      ctr,#:sqlp

square_ret    ret

''--------[sqrt]-----------------------------------------------------

'' acc = sqrt(acc)      (Square root routine from "Propeller Guts".)

sqrt          mov       accy,#0
              mov       accx,#0
              mov       ctr,#16
:rtlp         shl       acc,#1 wc

              rcl       accy,#1
              shl       acc,#1 wc
              rcl       accy,#1
              shl       accx,#2 
              or        accx,#1
              cmpsub    accy,accx wc
              shr       accx,#2
              rcl       accx,#1
              djnz      ctr,#:rtlp

              mov       acc,accx
sqrt_ret      ret

''-------------------------------------------------------------------

d_audio       long      0-0
d_ctra0       long      %10110 << 26 | if_pin << 9 | fb_pin
d_ctrb0       long      %10110 << 26 | qf_pin << 9 | fb_pin
sample        long      clks
half          long      mclks

acc           res       1
accx          res       1
accy          res       1
ctr           res       1
i_samp        res       1
q_samp        res       1
time          res       1

''===================================================================

'' Audio Output

              org       0                
audio_out     mov       ctra,a_ctra0            'Set counter for DUTY mode output.
              mov       dira,a_dira0

:aulp         rdlong    aud,a_audio             'Get the instantaneous audio amplitude.
              shl       aud,#20                 'Boost the volume to audible levels.
              mov       frqa,aud                'Make boosted value the DUTY level.
              jmp       #:aulp                  'Keep doing it.

a_audio       long      0-0
a_ctra0       long      %00110 << 26 | au_pin
a_dira0       long      1 << au_pin

aud           res       1

''-----------[ TERMS OF USE ]---------------------------------------------------
''
'' Permission is hereby granted, free of charge, to any person obtaining a copy of
'' this software and associated documentation files (the "Software"), to deal in
'' the Software without restriction, including without limitation the rights to use,
'' copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
'' Software, and to permit persons to whom the Software is furnished to do so,
'' subject to the following conditions: 
''
'' The above copyright notice and this permission notice shall be included in all
'' copies or substantial portions of the Software. 
''
'' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'' INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
'' PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
'' HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
'' OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  